fntenc.o fntenc.d : fntenc.c fntenc.h bytesex.h util.h enc/blackboard.h \
 enc/cork.h enc/ot1.h enc/ot1wofl.h enc/texmext.h enc/texmital.h \
 enc/texmsym.h enc/textt.h enc/eurosym.h enc/textcomp.h enc/marvo98.h \
 enc/marvo00.h enc/latexsym.h enc/eufrak.h enc/dummy.h enc/amssymba.h \
 enc/amssymbb.h enc/euex.h
